/*
 * Program: SSL-WOPR Server
 * Source : ssl-therad.cpp
 * Purpose: Implementation of OpenSSL thread callbacks.
 *
 * Copyright (c) 2009, 2010 Oliver Mueller.
 * All rights reserved.
 * http://www.cogito-ergo-sum.org
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the author be held liable for any damages arising from
 * the use of this software.
 *
 */

#include <exception>
#include <new>
#include <unistd.h>
#include "ssl-thread.h"

using namespace std;

pthread_mutex_t *ssl_mutex = NULL;

int ssl_init_mutex()
{
    if(ssl_mutex == NULL) {
        try {
            int numlocks = CRYPTO_num_locks(), ret = 0;
            ssl_mutex = new pthread_mutex_t[numlocks];
            for(int n = 0; n < numlocks; n++)
                if(pthread_mutex_init(&ssl_mutex[n], NULL))
                    ret = -2;
            return ret;
        } catch(bad_alloc&) {
            ssl_mutex = NULL;
            return -1;
        }
    } else
        return 1;
}

ZOS_C_DECL
void ssl_lock(int mode, int n, const char *file, int line)
{
    if(mode & CRYPTO_LOCK)
        pthread_mutex_lock(&ssl_mutex[n]);
    else
        pthread_mutex_unlock(&ssl_mutex[n]);
}

#if OPENSSL_VERSION_NUMBER >= 0x01000000f

// We are using OpenSSL 1.0.0 or higher.
// We will use the new CRYPTO_THREADID_* functions.

ZOS_C_DECL
void ssl_thread_id(CRYPTO_THREADID *id)
{
#ifdef PTHREADS_ON_ZOS
    pthread_t p = pthread_self();
    unsigned long x = 0L;
    int i;
    for(i = 0; i < 8; i++) {
        x <<= 8;
        x |= ((unsigned long) p.__[i] & 0x000000FFL);
    }
    CRYPTO_THREADID_set_numeric(id, x);
#else
    CRYPTO_THREADID_set_pointer(id, (void*) pthread_self());
#endif
}

#else // OpenSSL < 1.0.0

// We are using OpenSSL before 1.0.0.
// We have to use the old thread_id based functions
// which are using values of type unsigned long to
// identify threads.

ZOS_C_DECL
unsigned long ssl_thread_id()
{
#ifdef PTHREADS_ON_ZOS
    pthread_t p = pthread_self();
    unsigned long x = 0L;
    int i;
    for(i = 0; i < 8; i++) {
        x <<= 8;
        x |= ((unsigned long) p.__[i] & 0x000000FFL);
    }
    return x;
#else
    return (unsigned long) pthread_self();
#endif
}

#endif // OpenSSL == 1.0.0?

ZOS_C_DECL
struct CRYPTO_dynlock_value *ssl_dynlock_create(const char *file, int line)
{
    try {
        struct CRYPTO_dynlock_value *l = new CRYPTO_dynlock_value;
        if(pthread_mutex_init(&l->mutex, NULL)) {
            // Initialization of mutex failed.
            delete l;
            return NULL;
        } else
            return l;
    } catch(bad_alloc&) {
        return NULL;
    }
}

ZOS_C_DECL
void ssl_dynlock_destroy(struct CRYPTO_dynlock_value *l, const char *file, int line)
{
    pthread_mutex_destroy(&l->mutex);
    delete l;
}

ZOS_C_DECL
void ssl_dynlock_lock(int mode, struct CRYPTO_dynlock_value *l, const char *file, int line)
{
    if(mode & CRYPTO_LOCK)
        pthread_mutex_lock(&l->mutex);
    else
        pthread_mutex_unlock(&l->mutex);
}
